from archijson import ArchiServer
from sensitive_info import URL, TOKEN

server = ArchiServer(URL, TOKEN, 'engine')


def on_connect():
    print('exchanging')
    server.send('client', {'msg': 'hello'})


def on_receive(id, body):
    print(id)
    print(body)


server.on_connect = on_connect
server.on_receive = on_receive

# import socketio

# sio = socketio.Client()

# @sio.event
# def connect():
#     print('connection established')

# @sio.event
# def disconnect():
#     print('disconnected from server')

# sio.connect(URL)
# sio.wait()



from archijson.geometry import Cylinder

if __name__ == '__main__':
    c = Cylinder(5, 10, properties={'color': 'red'})
    print(c)
    print(type(c))
    print(c.data)

    print(c.validate)

    print(Cylinder.is_valid(c.data))




import socketio
import json


class ArchiServer:

    def __init__(self, URL, TOKEN, IDENTITY):
        self.URL = URL
        self.TOKEN = TOKEN
        self.IDENTITY = IDENTITY
        self.socket = self.setup()

    def setup(self):
        socket = socketio.Client(ssl_verify=False)

        @socket.event
        def connect():
            socket.emit('register', {
                        'token': self.TOKEN, 'identity': self.IDENTITY}, callback=self.__connect_callback)

        @socket.event
        def receive(ctx):
            if(type(ctx) == str):
                ctx = json.loads(ctx)
            id = ctx['id']
            body = ctx['body']

            self.on_receive(id, body)

        socket.connect(self.URL, transports='websocket')
        return socket

    def send(self, identity, body, id=None):
        # print('sending...')
        o = {'to': identity, 'body': body}
        if(id != None):
            o['id'] = id

        # print(o)
        self.socket.emit('exchange', o, callback=self.__print_status)

    def on_receive(self, id, body):
        pass

    def on_connect(self):
        pass

    def __connect_callback(self, o):
        self.__print_status(o)
        self.on_connect()

    def __print_status(self, o):
        print(o['status'])



from archijson.geometry import Circle
import archijson.geometry as geometry
import json


def test_circle():
    c = Circle(5, 10, properties={'color': 'red'})

    assert c.type == 'Circle'
    print(c.type)
    print(c.data)
    print(c.param)
    print(c.validate)
    assert c.is_valid(c.data)



from geometry import Vertices, Faces, Mesh
from itertools import chain

def to_mesh(vs, fs):
    assert (type(vs) == list) & (type(vs[0]) == list)
    assert (type(fs) == list) & (type(fs[0]) == list)

    v_size = len(vs[0])
    v_coords = list(chain(*vs))

    v = Vertices(v_coords, v_size)

    f_index = []
    f_count = []
    f_size = []

    mp = {}
    for id, face in enumerate(fs):
        try:
            mp[len(face)].append(id)
        except Exception:
            mp[len(face)] = [id]

    for size in mp:
        idx = mp[size]

        f_size.append(size)
        f_count.append(len(idx))

        for id in idx:
            f_index.extend(fs[id])

    f = Faces(f_count, f_size, f_index)
    return Mesh(v, f)




from archijson.geometry import Vertices, Segments, Faces, Mesh


def test_mesh():

    v = Vertices([1, 1, 1, 3, 3, 3], 2)

    print(v.validate)
    print(v.param)
    print(v.coordinates)
    print(Vertices.is_valid(v.data))
    print(v.toList())

    s = Segments([1, 1, 1, 3, 3, 3], 2)
    print(s.data)
    print(s.validate)
    print(s.toList())

    f = Faces([1], [3], [1, 2, 3])
    print(f.data)

    ff = Faces([1], [2], [1, 2])
    print(ff.validate)

    m = Mesh(v, f)
    print(m.data)
    print(m.validate)

    print('')





import schema
import json
from archijson.geometry import BaseGeometry
from archijson.helper import chunks



class Vertices(BaseGeometry):
    def __init__(self, coordinates, size=3, pointSize=10, **kwargs):
        if('type' in kwargs):
            del kwargs['type']
        super(Vertices, self).__init__('Vertices', **kwargs)
        self.size = size
        self.pointSize = pointSize
        self.coordinates = coordinates
    
    @staticmethod
    def DATASCHEMA():
        return schema.Schema({
            'type': 'Vertices',
            'size': schema.And(int, lambda x: x > 0),
            'coordinates': list
        }, ignore_extra_keys=True)
    
    @property
    def param(self):
        return {
            'type': 'Vertices',
            'size': self.size,
            'coordinates': self.coordinates
        }
    
    @property
    def data(self):
        return {
            **super(Vertices, self).data,
            **self.param
        }

    def toList(self):
        return list(chunks(self.coordinates, self.size))
    
    @property
    def validate(self):
        return {
            **super(Vertices, self).validate(),
            **Vertices.DATASCHEMA().validate(self.data)
        }

    @staticmethod
    def is_valid(obj):
        return BaseGeometry.DATASCHEMA().is_valid(obj) & Vertices.DATASCHEMA().is_valid(obj)



class Segments(BaseGeometry):
    def __init__(self, coordinates, size=3, closed=True, filled=False, **kwargs):
        if('type' in kwargs):
            del kwargs['type']
        super(Segments, self).__init__('Segments', **kwargs)
        self.coordinates = coordinates
        self.size = size
        self.closed = closed
        self.filled = filled

    @staticmethod
    def DATASCHEMA():
        return schema.Schema({
            'type': 'Segments',
            'size': schema.And(int, lambda x: x > 0),
            'closed': schema.Optional(bool),
            'filled': schema.Optional(bool),
            'coordinates': list
        }, ignore_extra_keys=True)
    
    @property
    def param(self):
        return {
            'closed': self.closed,
            'filled': self.filled,
            'size': self.size,
            'coordinates': self.coordinates
        }
    
    @property
    def data(self):
        return {
            **super(Segments, self).data,
            **self.param
        }
    
    @property
    def validate(self):
        return {
            **super(Segments, self).validate(),
            **Segments.DATASCHEMA().validate(self.data)
        }

    @staticmethod
    def is_valid(obj):
        return BaseGeometry.DATASCHEMA().is_valid(obj) & Segments.DATASCHEMA().is_valid(obj)
    
    def toList(self):
        return list(chunks(self.coordinates, self.size))
    


class Prism(BaseGeometry):
    def __init__(self, segments: Segments, height, **kwargs):
        if('type' in kwargs):
            del kwargs['type']
        super(Prism, self).__init__('Prism', **kwargs)
        self.segments = segments
        self.height = height


    @property
    def param(self):
        return {
            'segments': self.segments.data,
            'height': self.height,
        }

    @property
    def data(self):
        return {
            **super(Prism, self).data,
            **self.param
        }

    @property
    def validate(self):
        return {
            **super(Prism, self).validate(),
            'segments': self.segments.validate
        }

    @staticmethod
    def is_valid(obj):
        return BaseGeometry.DATASCHEMA().is_valid(obj) & Segments.DATASCHEMA().is_valid(obj['segments'])




class Faces:
    def __init__(self, count=[], size=[], index=[], properties={}, **kwargs):
        self.type = 'Faces'
        self.count = count
        self.size = size
        self.index = index
        self.properties = properties
    
    @staticmethod
    def DATASCHEMA():
        return schema.Schema({
            'type': 'Faces',
            'count': [int],
            'size': [int],
            'index': [int],
            schema.Optional('properties'): dict
        })

    @staticmethod
    def is_valid(obj): 
        return Faces.DATASCHEMA().is_valid(obj)

    @property
    def validate(self):
        return Faces.DATASCHEMA().validate(self.data)
    
    @property
    def data(self):
        return {
            'type': self.type,
            'count': self.count,
            'size': self.size,
            'index': self.index,
            'properties': self.properties,
        }
    
    def __str__(self):
        return ("{}: {}".format(self.type, json.dumps(self.data, indent = 4)))
    
    def toList(self):
        ret, lst = [], 0
        for i, cnt in enumerate(self.count):
            sz = self.size[i]
            ret.extend(list(chunks(self.index[lst:lst+sz*cnt], sz)))
            lst += sz * cnt
        return ret


class Mesh(BaseGeometry):
    def __init__(self, vertices, faces, **kwargs):
        if('type' in kwargs):
            del kwargs['type']
        super(Mesh, self).__init__('Mesh', **kwargs)
        self.vertices = vertices
        self.faces = faces
    
    @property
    def param(self):
        return {
            'vertices': self.vertices.param,
            'faces': self.faces.data,
        }
    
    @property
    def data(self):
        return {
            **super(Mesh, self).data,
            **self.param
        }



    @property
    def validate(self):
        return {
            **super(Mesh, self).validate(),
            'vertices': self.vertices.validate,
            'faces': self.faces.validate
        }

    @staticmethod
    def is_valid(obj):
        return BaseGeometry.DATASCHEMA().is_valid(obj) & Vertices.DATASCHEMA().is_valid(obj['vertices']) & Faces.DATASCHEMA().is_valid(obj['faces'])



from archijson.geometry import Cuboid
import archijson.geometry as geometry
import json


def test_cuboid():

    b = Cuboid(3, 3, 3)
    print(dir(b))
    b.w = 2
    print(b.w)
    print(b.data)
    print(b.matrix)
    print(b.validate)

    print(json.dumps(b.validate))

    nb = json.loads('{"type": "Cuboid", "uuid": "ba79302c-c3a3-407a-a684-fa31eef602d6", "matrix": [2.0, 0.0, 0.0, 0.0, 0.0, 3.0, 0.0, 0.0, 0.0, 0.0, 3.0, 0.0, 0.0, 0.0, 0.0, 1.0],  "position": {"x": 0, "y": 0, "z": 0}, "length": 2, "width": 3, "height": 3}')
    print(nb)
    print(Cuboid.is_valid(nb))

    cb = geometry.call['Cuboid'](**nb)
    # cb = Cuboid(1, 21, 23, type='Hello')
    # print(cb.data)

    


from archijson import ArchiServer, ArchiJSON
from archijson.geometry import BasePoint
from sensitive_info import URL, TOKEN

server = ArchiServer(URL, TOKEN, 'archijson')

def on_receive(id, body):
    print(id)
    print(body)
    archijson = ArchiJSON(body)
    print(archijson.geometries)

    o = ArchiJSON()
    
    for geom in archijson.geometries:
        p = geom.position
        
        geom.position = BasePoint(p.x, p.y - 900, p.z)
        geom.uuid = ''
        o.geometries.append(geom)
        
    server.send('client', o.toJSON(), id)

server.on_receive = on_receive



import sys
import os
print(sys.path)
sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
print(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
print(sys.path)




from archijson.geometry import Prism, Segments

def test_prism():

    s = Segments([1, 1, 1, 3, 3, 3], 2, closed=True)
    p = Prism(s, 10, properties={'color': 'blue'})

    assert p.type == 'Prism'
    print(p.type)
    print(p.data)
    print(p.param)
    print(p.validate)
    assert p.is_valid(p.data)

    print(p.data)




def identity(n):
    arr = [0. for i in range(n * n)]
    for i in range(n):
        arr[i * n + i] = 1.
    return arr


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]





from archijson.geometry.base import BasePoint, BaseGeometry
from archijson.geometry.shape import Cuboid, Cylinder, Plane, Circle, Sphere
from archijson.geometry.mesh import Vertices, Segments, Faces, Mesh, Prism

call = {
    'Cuboid': Cuboid,
    'Cylinder': Cylinder,
    'Plane': Plane,
    'Sphere': Sphere,
    'Circle': Circle,
    'Vertices': Vertices,
    'Segments': Segments,
    'Faces': Faces,
    'Mesh': Mesh,
    'Prism': Prism,
}




import schema, math
from archijson.geometry import BaseGeometry


class Cuboid(BaseGeometry):
    def __init__(self, length=1, width=1, height=1, **kwargs):
        if('type' in kwargs):
            del kwargs['type']
        super(Cuboid, self).__init__('Cuboid', **kwargs)
        self.__length = length
        self.__width = width
        self.__height = height
        self.matrix[0] = float(length)
        self.matrix[5] = float(width)
        self.matrix[10] = float(height)

    @property
    def length(self):
        return self.__length

    @property
    def width(self):
        return self.__width

    @property
    def height(self):
        return self.__height

    @length.setter
    def length(self, length):
        m = self.matrix[0]
        self.matrix[0] = m*length/self.length
        self.__w = length

    @width.setter
    def width(self, width):
        m = self.matrix[5]
        self.matrix[5] = m*width/self.width
        self.__width = width

    @height.setter
    def height(self, height):
        m = self.matrix[10]
        self.matrix[10] = m*height/self.height
        self.__height = height

    @property
    def param(self):
        return {
            'length': self.length,
            'width': self.width,
            'height': self.height
        }

    @property
    def data(self):
        return {
            **super(Cuboid, self).data,
            **self.param
        }

    def DATASCHEMA():
        return schema.Schema({
            'type': 'Cuboid',
            'length': schema.And(schema.Or(float, int), lambda x: x > 0),
            'width': schema.And(schema.Or(float, int), lambda x: x > 0),
            'height': schema.And(schema.Or(float, int), lambda x: x > 0),
        }, ignore_extra_keys=True)

    @property
    def validate(self):
        a = super(Cuboid, self).validate()
        b = Cuboid.DATASCHEMA().validate(self.data)
        return {
            **a,
            **b
        }

    def is_valid(obj):
        return BaseGeometry.DATASCHEMA().is_valid(obj) & Cuboid.DATASCHEMA().is_valid(obj)


class Plane(BaseGeometry):

    def __init__(self, length=1, width=1, **kwargs):
        if('type' in kwargs):
            del kwargs['type']
        super(Plane, self).__init__('Plane', **kwargs)
        self.__length = length
        self.__width = width
        self.matrix[0] = float(length)
        self.matrix[5] = float(width)

    @property
    def length(self):
        return self.__length
    
    @property
    def width(self):
        return self.__width
    
    @length.setter
    def length(self, length):
        m = self.matrix[0]
        self.matrix[0] = m*length/self.length
        self.__length = length
    
    @width.setter
    def width(self, width):
        m = self.matrix[5]
        self.matrix[5] = m*width/self.width
        self.__width = width
    
    @property
    def param(self):
        return {
            'length': self.length,
            'width': self.width
        }

    @property
    def data(self):
        return {
            **super(Plane, self).data,
            **self.param
        }
    
    def DATASCHEMA():
        return schema.Schema({
            'type':'Plane',
            'length': schema.And(schema.Or(float, int), lambda x: x > 0),
            'width': schema.And(schema.Or(float, int), lambda x: x > 0)
        }, ignore_extra_keys=True)
    
    @property
    def validate(self):
        return {
            **super(Plane, self).validate(),
            **Plane.DATASCHEMA().validate(self.data)
        }
    
    def is_valid(obj):
        return BaseGeometry.DATASCHEMA().is_valid(obj) & Plane.DATASCHEMA().is_valid(obj)




class Cylinder(BaseGeometry):
    def __init__(self, radius=1, height=1, **kwargs):
        if('type' in kwargs):
            del kwargs['type']
        super(Cylinder, self).__init__('Cylinder', **kwargs)

        self.__radius = radius
        self.__height = height
        self.matrix[0] = float(radius)
        self.matrix[5] = float(radius)
        self.matrix[10] = float(height)

    @property
    def radius(self):
        return self.__radius

    @property
    def height(self):
        return self.__height

    @radius.setter
    def radius(self, radius):
        m = self.matrix[0]
        self.matrix[0] = m*radius/self.radius
        self.matrix[5] = m*radius/self.radius
        self.__r = radius

    @height.setter
    def height(self, height):
        m = self.matrix[10]
        self.matrix[10] = m*height/self.height
        self.__height = height

    @property
    def param(self):
        return {
            'radius': self.radius,
            'height': self.height
        }

    @property
    def data(self):
        return {
            **super(Cylinder, self).data,
            **self.param
        }

    
    def DATASCHEMA():
        return schema.Schema({
            'type': 'Cylinder',
            'radius': schema.And(schema.Or(float, int), lambda x: x > 0),
            'height': schema.And(schema.Or(float, int), lambda x: x > 0)
        }, ignore_extra_keys=True)

    @property
    def validate(self):
        return {
            **super(Cylinder, self).validate(),
            **Cylinder.DATASCHEMA().validate(self.data)
        }
    
    def is_valid(obj):
        return BaseGeometry.DATASCHEMA().is_valid(obj) & Cylinder.DATASCHEMA().is_valid(obj)


class Circle(BaseGeometry):
    def __init__(self, radius=1, theta=math.pi, **kwargs):
        if('type' in kwargs):
            del kwargs['type']
        super(Circle, self).__init__('Circle', **kwargs)
        self.__radius = radius
        self.__theta = theta
        self.matrix[0] = float(radius)
        self.matrix[5] = float(radius)

    def __str__(self):
        return ("Circle: radius = {}, theta = {}".format(self.radius, self.theta))

    @property
    def radius(self):
        return self.__radius

    @property
    def theta(self):
        return self.__theta

    @radius.setter
    def radius(self, radius):
        m = self.matrix[0]
        self.matrix[0] = m*radius/self.radius
        self.__radius = radius

    @theta.setter
    def theta(self, theta):
        m = self.matrix[5]
        self.matrix[5] = m*theta/self.theta
        self.__theta = theta

    @property
    def param(self):
        return {
            'radius': self.radius,
            'theta': self.theta
        }

    @property
    def data(self):
        return {
            **super(Circle, self).data,
            **self.param
        }

    def DATASCHEMA():
        return schema.Schema({
            'type': 'Circle',
            'radius': schema.And(schema.Or(float, int), lambda x: x > 0),
            'theta': schema.And(schema.Or(float, int), lambda x: x > 0)
        }, ignore_extra_keys=True)

    @property
    def validate(self):
        return {
            **super(Circle, self).validate(),
            **Circle.DATASCHEMA().validate(self.data)
        }

    @staticmethod
    def is_valid(obj):
        return BaseGeometry.DATASCHEMA().is_valid(obj) & Circle.DATASCHEMA().is_valid(obj)


class Sphere(BaseGeometry):
    def __init__(self, radius=1, **kwargs):
        if('type' in kwargs):
            del kwargs['type']
        super(Sphere, self).__init__('Sphere', **kwargs)
        self.__radius = radius
        self.matrix[0] = float(radius)
        self.matrix[5] = float(radius)
        self.matrix[10] = float(radius)

    @property
    def radius(self):
        return self.__radius

    @radius.setter
    def radius(self, radius):
        m = self.matrix[0]
        self.matrix[0] = m*radius/self.radius
        self.matrix[5] = m*radius/self.radius
        self.matrix[10] = m*radius/self.radius
        self.__radius = radius

    @property
    def param(self):
        return {
            'radius': self.radius
        }

    @property
    def data(self):
        return {
            **super(Sphere, self).data,
            **self.param
        }

    def DATASCHEMA():
        return schema.Schema({
            'type': 'Sphere',
            'radius': schema.And(schema.Or(float, int), lambda x: x > 0)
        }, ignore_extra_keys=True)

    @property
    def validate(self):
        return {
            **super(Sphere, self).validate(),
            **Sphere.DATASCHEMA().validate(self.data)
        }

    @staticmethod
    def is_valid(obj):
        return BaseGeometry.DATASCHEMA().is_valid(obj) & Sphere.DATASCHEMA().is_valid(obj)


from archijson.geometry import BasePoint

def test_create():
    p1 = BasePoint(1, 2, 3)
    print(p1)
    assert p1.x == 1
    assert p1.y == 2
    assert p1.z == 3

    p2 = BasePoint([1, 1, 1])
    print(p2)

    p3 = BasePoint({'x': 909.1648356554529, 'y': 51.31071849793332, 'z': 0})
    print(p3)

    assert type(p1) == BasePoint

    p4 = BasePoint(p2)
    assert p4!=p2




import json
import archijson.geometry as geometry

class ArchiJSON:
    def __init__(self, data=None):
        if(type(data) == str):
            data = json.loads(data)

        try:
            self.properties = data['properties']
            self.geometryElements = data['geometryElements']
        except Exception:
            self.properties = {}
            self.geometryElements = []

        if(type(self.properties) == str):
            self.properties = json.loads(self.properties)
        if(type(self.geometryElements) == str):
            self.geometryElements = json.loads(self.geometryElements)

        self.parse_geometry()
    

    def parse_geometry(self):
        self.geometries = []
        for el in self.geometryElements:
            geom = self.__from_element(el)
            self.geometries.append(geom)


    def __from_element(self, obj):
        if(obj['type'] == 'Mesh'):
            v = geometry.call['Vertices'](**obj['vertices'])
            f = geometry.call['Faces'](**obj['faces'])
            del obj['faces']
            del obj['vertices']
            return geometry.call['Mesh'](v, f, **obj)

        elif (geometry.call[obj['type']].is_valid(obj)):
            return geometry.call[obj['type']](**obj)
    
    
    @property
    def data(self):
        return {
            'properties': self.properties,
            'geometries': self.geometries
        }
    
    def toJSON(self):
        for geom in self.geometries:
            if(geometry.call[geom.type].is_valid(geom.data)):
                self.geometryElements.append(geom.data)
            else:
                raise ValueError('Invalid geometry')
        return {
            'properties': self.properties,
            'geometryElements': self.geometryElements
        }
        
        
    def __str__(self):
        return "ArchiJSON: {}".format(json.dumps(self.data, indent = 4))





from archijson.geometry import BasePoint, BaseGeometry
import numpy as np

import pytest

def test_basepoint():
    pt = BasePoint(0, 3, 5)
    assert pt.x == 0
    assert pt.y == 3
    assert pt.z == 5
    print(pt.data)
    
    testpt = {'x': 0, 'y': 2, 'z': 3.1}
    print(BasePoint.DATASCHEMA().validate(testpt))

def test_basegeometry():
    bg = BaseGeometry('Geom')
    print(bg)
    print(bg.data)
    print(bg.validate())

    bg.position = BasePoint(2, 3, 4)
    print(bg.data)


test_basepoint()
False

test_basegeometry()







from setuptools import setup, find_packages
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='archijson', 
    version='0.1.3',
    packages=find_packages(),
    install_requires=['schema', 'python-socketio', 'requests', 'websocket-client', 'trimesh', 'mapbox-earcut', 'numpy'],
    description='ArchiJSON: A Light Weight Web Data Exchange Format for Architectrual Design',
    license='GNU General Public License v3.0',
    url='https://github.com/Inst-AAA/archijson',
    author='Yichen Mo',
    author_email='moyichen@seu.edu.cn',
    long_description=long_description,
    long_description_content_type = 'text/markdown',
    keywords=['architecture', 'engineering', 'design', 'json', 'exchange format'],

    project_urls={
        "ArchiWeb": "https://web.archialgo.com",
        "Documentation": "https://docs.web.archialgo.com",
        "Issues": "https://github.com/Inst-AAA/archijson/issues",
        "Repository": "https://github.com/Inst-AAA/archijson",
    },
        classifiers=[
        # https://pypi.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows'
    ],
)



from archijson.geometry import Plane

def test():
    c = Plane(5, 10, properties={'color': 'red'})
    print(c)
    print(type(c))
    print(c.data)

    print(c.validate)

    print(Plane.is_valid(c.data))




import uuid
import schema
import json
from helper import identity


class BasePoint:

    def DATASCHEMA():
        return schema.Schema({
            'x': schema.Or(float, int),
            'y': schema.Or(float, int),
            'z': schema.Or(float, int)
        })

    def __init__(self, x=0, y=0, z=0):
        if((type(x) == dict) and BasePoint.is_valid(x)):
            self.x = x['x']
            self.y = x['y']
            self.z = x['z']
        elif((type(x) == tuple) or (type(x) == list)):
            self.x = x[0]
            self.y = x[1]
            self.z = x[2]
        elif((type(x) == BasePoint) and BasePoint.is_valid(x.data)):
            self.x = x.x
            self.y = x.y
            self.z = x.z
        else:
            self.x = float(x)
            self.y = float(y)
            self.z = float(z)

    def __str__(self):
        return ("Point: x = {}, y = {}, z = {}".format(self.x, self.y, self.z))

    def validate(self):
        return BasePoint.DATASCHEMA().validate(self.data)

    def is_valid(obj):
        return BasePoint.DATASCHEMA().is_valid(obj)


    @property
    def data(self):
        return {
            'x': self.x,
            'y': self.y,
            'z': self.z
        }

class BaseGeometry:

    def DATASCHEMA():
        return schema.Schema({
            'type': str,
            'uuid': str,
            'matrix': schema.And(list, lambda x: len(x) == 16),
            schema.Optional('properties'): dict,
            'position': BasePoint.DATASCHEMA(),
        }, ignore_extra_keys=True)

    def __init__(self, type, uuid=str(uuid.uuid4()), matrix=identity(4), properties={}, position=BasePoint(),  **kwargs):
        self.type = type
        self.uuid = uuid
        self.matrix = matrix
        self.properties = properties
        self.__position = BasePoint(position)

    @property
    def position(self):
        return self.__position

    @position.setter
    def position(self, position):
        p = BasePoint(position)

        self.matrix[12] = p.x
        self.matrix[13] = p.y
        self.matrix[14] = p.z
        self.__position = p

    def __str__(self):
        return ("{}: {}".format(self.type, json.dumps(self.data, indent = 4)))

    def validate(self):
        return BaseGeometry.DATASCHEMA().validate(self.data)

    @property
    def data(self):
        return {
            'type': self.type,
            'uuid': self.uuid,
            'matrix': self.matrix,
            'properties': self.properties,
            'position': self.position.data,
        }




class A:
    def deco(self, text):
        def docorator(func):
            def wrapper(*args, **kwargs):
                print('{}: {}'.format(text, func.__name__))
                return func(*args, **kwargs)
            return wrapper
        return docorator


    def fly(self):
        print('fly')

    def behavior(self):
        self.fly()

def test():
    a = A()
    a.fly()
    a.behavior()
    print(a.fly)

    @a.deco('override')
    def fly():
        print('no fly')

    a.fly = fly

    a.fly()
    a.behavior()

